import React, { useRef, useMemo, useEffect } from 'react';
import { render } from 'react-nil';
import zcreate from 'zustand';

function useImmediateEffect(effect, deps) {
  var cleanup = useRef();
  useMemo(function () {
    if (cleanup.current) {
      cleanup.current();
    }

    cleanup.current = effect();
  }, deps);
}
var serializeHookArgs = function serializeHookArgs(hookArgs) {
  try {
    return stableStringify(hookArgs);
  } catch (_unused) {
    throw new Error('Invalid Hook Args');
  }
};

function stableStringify(value) {
  return JSON.stringify(value, stableStringifyReplacer);
}

function stableStringifyReplacer(_key, value) {
  if (typeof value === 'function') {
    throw new Error();
  }

  if (isPlainObject(value)) {
    return Object.keys(value).sort().reduce(function (result, key) {
      result[key] = value[key];
      return result;
    }, {});
  }

  return value;
}

function isPlainObject(o) {
  if (!hasObjectPrototype(o)) {
    return false;
  } // If has modified constructor


  var ctor = o.constructor;

  if (typeof ctor === 'undefined') {
    return true;
  } // If has modified prototype


  var prot = ctor.prototype;

  if (!hasObjectPrototype(prot)) {
    return false;
  } // If constructor does not have an Object-specific method


  if (!prot.hasOwnProperty('isPrototypeOf')) {
    return false;
  } // Most likely a plain Object


  return true;
}

function hasObjectPrototype(o) {
  return Object.prototype.toString.call(o) === '[object Object]';
}

function usePrevious(value) {
  // The ref object is a generic container whose current property is mutable ...
  // ... and can hold any value, similar to an instance property on a class
  var ref = useRef(); // Store current value in ref

  useEffect(function () {
    ref.current = value;
  }, [value]); // Only re-run if value changes
  // Return previous value (happens before update in useEffect above)

  return ref.current;
}

var debug = function debug() {
  var _console;

  return  (_console = console).log.apply(_console, arguments) ;
};

var SET_STATE_ERR = 'Zusteller stores do not support setState. You must set state via methods returned from your hook.';
function create(useHook, defaultIsEqual) {
  var storeFamily = {};

  function makeInternalSelector(selector, id) {
    if (id === void 0) {
      id = 0;
    }

    return function (_ref) {
      var value = _ref.value;
      debug('makeInternalSelector', id, value);
      if (value == null) return value;

      if (typeof selector === 'function') {
        return selector(value);
      }

      return value;
    };
  }

  function makeInternalIsEqual(eqFn, i) {
    if (typeof eqFn === 'function') {
      return function (_ref2, _ref3) {
        var prevValue = _ref2.value;
        var nextValue = _ref3.value;
        return eqFn(prevValue, nextValue);
      };
    }

    return undefined;
  }

  var subscriberCounts = {}; // Make the default no-args store

  var defaultStore = makeNewZustandStore('[]', [], 0); // const defaultStore = storeFamily['[]'];

  function makeNewZustandStore(hookArgsHash, hookArgs, id) {
    var useZStore = zcreate(function () {
      return {
        value: undefined
      };
    });

    var useStoreWrapper = function useStoreWrapper() {
      return useZStore.apply(void 0, arguments);
    };

    useStoreWrapper.hash = hookArgsHash;

    useStoreWrapper.setState = function () {
      throw new Error(SET_STATE_ERR);
    };

    useStoreWrapper.getState = function () {
      debug('useStoreWrapper.getState', useZStore.getState().value);
      return useZStore.getState().value;
    };

    useStoreWrapper.subscribe = function (listener, selector, isEqual) {
      return useZStore.subscribe(listener, makeInternalSelector(selector, id), makeInternalIsEqual(isEqual));
    };

    useStoreWrapper.destroy = function () {
      return useZStore.destroy();
    };

    storeFamily[hookArgsHash] = useStoreWrapper;
    console.log(id, 'render');
    render( /*#__PURE__*/React.createElement(StoreComponent, {
      useHook: useHook,
      hookArgs: hookArgs,
      useZStore: useZStore
    }), {});
    return useZStore;
  }

  function useStore() {
    var hookArgs = [];
    var selector;
    var isEqual;

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    if (Array.isArray(args[0])) {
      hookArgs = args[0];
      selector = args[1];
      isEqual = args[2];
    } else if (args[0] != null && typeof args[0] != 'function') {
      throw new Error('Invalid Argument: Only a selector function or args array is supported. Did you mean to wrap your args in an array?');
    } else {
      selector = args[0];
      isEqual = args[1];
    } // Hook Args Hash
    // eslint-disable-next-line react-hooks/exhaustive-deps


    var hookArgsHash = useMemo(function () {
      return serializeHookArgs(hookArgs);
    }, hookArgs); // Register a subscribing component by incrementing
    // the subscriber subscriberCount

    var subscriberIndexRef = useRef(-1);
    useImmediateEffect(function () {
      if (subscriberCounts[hookArgsHash] == null) {
        subscriberCounts[hookArgsHash] = 0;
      }

      subscriberCounts[hookArgsHash] = subscriberCounts[hookArgsHash] + 1;
      subscriberIndexRef.current = subscriberCounts[hookArgsHash];
      return function () {
        subscriberCounts[hookArgsHash] = subscriberCounts[hookArgsHash] - 1;
        subscriberIndexRef.current = -1;
      };
    }, []);
    var id = subscriberIndexRef.current;
    debug(id, '-----------------');
    debug(id, 'hash', hookArgsHash);
    var prevHash = usePrevious(hookArgsHash);
    if (prevHash !== hookArgsHash) debug('hash changed', {
      prev: prevHash,
      next: hookArgsHash
    });
    debug(id, 'total subscribers', subscriberCounts[hookArgsHash]); // Set isEqual

    if (isEqual == null && defaultIsEqual) {
      isEqual = defaultIsEqual;
    } // Create or Get Cached Zustand Store Hook


    var isNewlyMadeRef = useRef(false);
    var useZStore = useMemo(function () {
      debug('enter memo');
      var cachedStore = storeFamily[hookArgsHash];

      if (cachedStore) {
        debug(id, 'use cached');
        return cachedStore;
      }

      debug(id, 'make new');
      isNewlyMadeRef.current = true;
      return makeNewZustandStore(hookArgsHash, hookArgs, id);
    }, [hookArgsHash]);
    var internalSelector = makeInternalSelector(selector, id);
    var internalIsEqual = makeInternalIsEqual(isEqual);
    var storeResult = useZStore(internalSelector, internalIsEqual);
    debug(id, 'hookResult', storeResult);
    var defaultResult = internalSelector(defaultStore.getState());
    debug(id, 'defaultResult', defaultResult);

    if (isNewlyMadeRef.current && storeResult === undefined && defaultResult !== undefined) {
      debug(id, 'return defaultResult for newly made hook args store', defaultResult);
      isNewlyMadeRef.current = false;
      return defaultResult;
    }

    return storeResult;
  }

  useStore.family = storeFamily;

  useStore.fromArgs = function (hookArgs) {
    if (hookArgs === void 0) {
      hookArgs = [];
    }

    var hookArgsHash = serializeHookArgs(hookArgs);
    debug('fromArgs', storeFamily[hookArgsHash]);
    var cachedStore = storeFamily[hookArgsHash];
    if (cachedStore) return cachedStore;
    makeNewZustandStore(hookArgsHash, hookArgs);
    return storeFamily[hookArgsHash];
  };

  useStore.getState = function (hookArgs) {
    if (hookArgs === void 0) {
      hookArgs = [];
    }

    return useStore.fromArgs(hookArgs).getState();
  };

  useStore.setState = function () {
    throw new Error(SET_STATE_ERR);
  };

  useStore.subscribe = function () {
    var hookArgs = [];
    var listener;
    var selector;
    var isEqual;

    for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
      args[_key2] = arguments[_key2];
    }

    if (Array.isArray(args[0])) {
      hookArgs = args[0];
      listener = args[1];
      selector = args[2];
      isEqual = args[3];
    } else {
      listener = args[0];
      selector = args[1];
      isEqual = args[2];
    }

    useStore.fromArgs(hookArgs).getState().subscribe(listener, makeInternalSelector(selector), makeInternalIsEqual(isEqual));
  };

  useStore.destroy = function (hookArgs) {
    if (hookArgs === void 0) {
      hookArgs = [];
    }

    return useStore.fromArgs(hookArgs).destroy();
  };

  return useStore;
}

function StoreComponent(_ref4) {
  var useHook = _ref4.useHook,
      _ref4$hookArgs = _ref4.hookArgs,
      hookArgs = _ref4$hookArgs === void 0 ? [] : _ref4$hookArgs,
      useZStore = _ref4.useZStore;
  debug('StoreComponent hookArgs', hookArgs);
  var result = useHook.apply(void 0, hookArgs);
  useZStore.getState().value = result;
  debug('StoreComponent render', result);
  useImmediateEffect(function () {
    debug('StoreComponent effect', result);
    useZStore.setState({
      value: result
    });
  }, [useZStore, result]);
  return null;
}

export default create;
//# sourceMappingURL=zusteller.esm.js.map
